/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#include <sstream>
#include <algorithm>
#include <cctype> // isalpha

#include "coeff_int.h"
#include "coeff_helper.h" // mod_mul, mod_inv

namespace pyred
{

// static members
bool Coeff_int::auto_vars{true};
std::string Coeff_int::set_to_one{};
uint64_t Coeff_int::prime = primes().front();
std::vector<std::pair<std::string,uint64_t>> Coeff_int::randvars{};

void Coeff_int::init(const std::vector<std::string>& vars, uint32_t n,
                     const std::string& set1) {
  if (n < 1 || n > primes().size()) {
    std::ostringstream ss;
    ss << "Invalid choice of prime: " << n << std::endl
       << "Must be 1<=n<=" << primes().size() << std::endl;
    throw input_error(ss.str());
  }
  if (!vars.empty()) {
    // If at lease one variable was set manually,
    // turn off auto variable declaration.
    auto_vars = false;
    if (!set1.empty() &&
        std::find(vars.cbegin(), vars.cend(), set1) == vars.cend()) {
      // If set1 is set and auto varible declaration is off,
      // set1 must be a declared variable.
      std::ostringstream ss;
      ss << "'" << set1 << "' is not declared as a variable." << std::endl;
      throw input_error(ss.str());
    }
  }
  set_to_one = set1;
  prime = primes().at(n-1);
  randvars.clear();
  randvars.reserve(vars.size());
  for (const auto& var: vars) {
    if (var == set1) {
      randvars.push_back({var, 1});
    }
    else {
      randvars.push_back({var, random_minmax(randvar_lbound(), prime)});
    }
  }
  return;
}

Coeff_int::Coeff_int() {}
Coeff_int::Coeff_int(const Coeff_int& a): c(a.c) {}

Coeff_int::Coeff_int(const std::string& s) {
  for (const auto& vc: randvars) {
    if (vc.first == s) {
      c = vc.second;
      return;
    }
  }
  std::istringstream ss{s};
  auto success = static_cast<bool>(ss >> c);
  if (!(success && ss.rdbuf()->in_avail() == 0)) {
    if (s.empty()) {
      // (ss >> c) fails if ss is empty
      throw parser_error("Coeff_int(): empty argument");
    }
    else if (std::isalpha(s.front())) {
      if (!auto_vars) {
        throw parser_error(
          std::string("invalid coefficient string \"") + s + "\"");
      }
      // auto_vars: add new symbols automatically and assign values to them.
      if (!is_valid_varname(s)) {
        throw parser_error(s + std::string(
          "; symbols are expected to be of the form [A-Za-z_][A-Za-z0-9_]*."));
      }
      if (s == set_to_one) {
        c = 1;
      }
      else {
        c = random_minmax(randvar_lbound(), prime);
      }
      randvars.push_back({s, c});
      // std::cout << "define symbol " << s << std::endl;
    }
    else {
      c = parse_longint(s, prime);
    }
  }
  else if (c >= prime) {
    // special case: the parsed value fits into c, but is >= prime
    c %= prime;
  }
}

Coeff_int& Coeff_int::operator+=(const Coeff_int& a) {
  c += a.c;
  if (c >= prime) c -= prime;
  return *this;
}
Coeff_int& Coeff_int::operator-=(const Coeff_int& a) {
  if (a.c > c) c += prime;
  c -= a.c;
  return *this;
}
Coeff_int& Coeff_int::operator*=(const Coeff_int& a) {
  c = mod_mul(c, a.c, prime);
  return *this;
}
Coeff_int& Coeff_int::operator/=(const Coeff_int& a) {
  c = mod_mul(c, mod_inv(a.c, prime), prime);
  return *this;
}
Coeff_int pow(const Coeff_int& b, const Coeff_int& a) {
  // assume that exponents are small integers
  Coeff_int result;
  uint64_t exp;
  uint64_t base;
  if (a.c == 0) {
    result.c = 1;
  }
  else {
    if (2*a.c < Coeff_int::prime) {
      // treat as positive exponent
      exp = a.c;
      base = b.c;
      result.c = base;
    }
    else {
      // treat as negative exponent
      exp = Coeff_int::prime - a.c;
      base = mod_inv(b.c, Coeff_int::prime); // =1/b.c
      result.c = base;
    }
    for (uint64_t i = 1; i != exp; ++i) {
      result.c = mod_mul(result.c, base, Coeff_int::prime);
    }
  }
  return result;
}

Coeff_int operator-(const Coeff_int& a) {
  if (a.c == 0) {
    return Coeff_int(a);
  }
  else {
    return Coeff_int(Coeff_int::prime - a.c);
  }
}
Coeff_int operator+(const Coeff_int& a, const Coeff_int& b) {
  auto sum = a.c + b.c;
  if (sum >= Coeff_int::prime) sum -= Coeff_int::prime;
  return Coeff_int(sum);
}
Coeff_int operator-(const Coeff_int& a, const Coeff_int& b) {
  auto diff = a.c;
  if (b.c > diff) diff += Coeff_int::prime;
  diff -= b.c;
  return Coeff_int(diff);
}
Coeff_int operator*(const Coeff_int& a, const Coeff_int& b) {
  return Coeff_int(mod_mul(a.c, b.c, Coeff_int::prime));
}
Coeff_int operator/(const Coeff_int& a, const Coeff_int& b) {
  return Coeff_int(
    mod_mul(a.c, mod_inv(b.c, Coeff_int::prime), Coeff_int::prime)
  );
}

bool operator==(const Coeff_int& a, const Coeff_int& b) {
  return (a.c == b.c);
}

bool operator!=(const Coeff_int& a, const Coeff_int& b) {
  return (a.c != b.c);
}

std::ostream& operator<<(std::ostream& out, const Coeff_int& a) {
  out << a.c;
  return out;
}

}
